/* apache includes */
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_core.h"
#include "http_log.h"

#include <openssl/x509.h>
#include <openssl/ssl.h>

#define EMAIL_SIZE    100 
#define CLIENTID_SIZE 9

typedef struct {
    char *prog;
    char *x509_cert_dir;
    char *x509_cert_file;
    char *x509_user_cert;
    char *x509_user_key;
    char *x509_user_proxy_prefix;
    char *tkt_cache_prefix;
} GridSrvConfigRec;

extern module MODULE_VAR_EXPORT grid_proxy_init_module;
 
#define cfgMerge(el,unset)  new->el = add->el == unset ? base->el : add->el
#define mySrvConfig(srv) (GridSrvConfigRec *)ap_get_module_config(srv->module_config,  &grid_proxy_init_module)

static int get_client_identity(pool *p, X509 *, char **);
static int get_user_proxy(char **, request_rec *);

void *grid_config_server_create(pool *p, server_rec *s) {
  GridSrvConfigRec *sc;

  sc = ap_palloc(p, sizeof(GridSrvConfigRec));
  sc->prog = NULL;
  sc->x509_cert_dir = NULL;
  sc->x509_cert_file = NULL;
  sc->x509_user_cert = NULL;
  sc->x509_user_key = NULL;
  sc->tkt_cache_prefix = NULL;
  sc->x509_user_proxy_prefix = NULL;

  return sc;
}

void *grid_config_server_merge(pool *p, void *basev, void *addv) {
   GridSrvConfigRec *base = (GridSrvConfigRec *)basev;
   GridSrvConfigRec *add = (GridSrvConfigRec *)add;
   GridSrvConfigRec *new = (GridSrvConfigRec *)ap_palloc(p, sizeof(GridSrvConfigRec));
 
   cfgMerge(prog, NULL);
   cfgMerge(x509_cert_dir, NULL);
   cfgMerge(x509_cert_file, NULL);
   cfgMerge(x509_user_cert, NULL);
   cfgMerge(x509_user_key, NULL);
   cfgMerge(x509_user_proxy_prefix, NULL);
   cfgMerge(tkt_cache_prefix, NULL);
 
   return new;
}

const char *grid_cmd_GridProxyInit(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server);
  struct stat sb;

  sc->prog = ap_server_root_relative(cmd->pool, arg); 
  if((sc->prog == NULL) || (stat(sc->prog, &sb) != 0) || 
     !S_ISREG(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "GridProxyInit: file '", sc->prog,
                      "' not exists or empty", NULL);
  return NULL;
}

const char *grid_cmd_X509CertDir(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server);
  struct stat sb;

  sc->x509_cert_dir = ap_server_root_relative(cmd->pool, arg);
  if((sc->x509_cert_dir == NULL) || (stat(sc->x509_cert_dir, &sb) != 0) ||
     !S_ISDIR(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "X509CertDir: file '", sc->x509_cert_dir,
                      "' not exists or empty", NULL);
  return NULL;                                                                  
 
}

const char *grid_cmd_X509CertFile(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server); 
  struct stat sb;
 
  sc->x509_cert_file = ap_server_root_relative(cmd->pool, arg);
  if((sc->x509_cert_file == NULL) || (stat(sc->x509_cert_file, &sb) != 0) ||
     !S_ISREG(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "X509CertFile: file '", sc->x509_cert_file,
                      "' not exists or empty", NULL);
  return NULL;
}

const char *grid_cmd_X509UserCert(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);

  sc->x509_user_cert = ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_cert, arg, size);

  return NULL;
}

const char *grid_cmd_X509UserKey(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server); 
  int size = strlen(arg);

  sc->x509_user_key=ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_key, arg, size);

  return NULL;
}

const char *grid_cmd_X509UserProxy(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);
  
  sc->x509_user_proxy_prefix=ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_proxy_prefix, arg, size);

  return NULL;
}

const char *grid_cmd_GridTktCache(cmd_parms *cmd, void *mconfig, char *arg) {
  GridSrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);
 
  sc->tkt_cache_prefix=ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->tkt_cache_prefix, arg, size);
 
  return NULL;
}                                                                                
static int get_client_identity(pool *p, X509 *cert, char **name) {
 
  X509_NAME *client=NULL;
  X509_NAME_ENTRY *entry=NULL;
  int i, n, entries;
  char *tmp;
 
  tmp = (unsigned char *)ap_pcalloc(p, EMAIL_SIZE);
  (*name) = (unsigned char *)ap_pcalloc(p, CLIENTID_SIZE);
 
  client = X509_get_subject_name(cert);
  X509_NAME_get_text_by_NID(client, NID_pkcs9_emailAddress, tmp,
                            EMAIL_SIZE);
  sscanf(tmp, "%[^@]", (*name));
  (*name)[strlen(*name)] = '\0';
  return OK;
}
  
static int get_arguments(char **x509_user_proxy, char **tkt_ccache, 
                         request_rec *r) {
  char *client_id;
  SSL *s = ap_ctx_get(r->connection->client->ctx, "ssl");
  GridSrvConfigRec *sc = mySrvConfig(r->server);                              

  if(s->session->peer == NULL) {
    fprintf(stderr, "unable to get ssl cert\n");
    return 0;
  }

  if(get_client_identity(r->pool, s->session->peer, &client_id) < 0) {
    fprintf(stderr, "wasn't able to get client's identity from the cert\n");
    return 0;
  }

  fprintf(stderr, "client id = %s\n", client_id);

  (*x509_user_proxy) = (char *)
     ap_pcalloc(r->pool, strlen(client_id)+strlen(sc->x509_user_proxy_prefix)+1);
  sprintf((*x509_user_proxy), "%s%s", sc->x509_user_proxy_prefix, client_id);

  (*tkt_ccache) = (char *)
     ap_pcalloc(r->pool, strlen(sc->tkt_cache_prefix)+strlen(client_id)+1);
  sprintf((*tkt_ccache), "%s%s", sc->tkt_cache_prefix, client_id);

  fprintf(stderr, "%s %s\n", (*x509_user_proxy), (*tkt_ccache));

  return 1;
}


static int grid_proxy_init_handler(request_rec *r) {

  int pid, status = 0;	
  char *x509_user_proxy, *tkt_ccache;
  SSL *s = ap_ctx_get(r->connection->client->ctx, "ssl");
  GridSrvConfigRec *sc = mySrvConfig(r->server);

  if(!get_arguments(&x509_user_proxy, &tkt_ccache, r)) 
    return HTTP_INTERNAL_SERVER_ERROR;

  if((pid = fork()) < 0) {
    fprintf(stderr, "fork failed\n");
  } else if(pid == 0) { /*child */
    fprintf(stderr, "%d: child process calling grid_proxy_init\n", getpid()); 
    if(execl(sc->prog, sc->prog, sc->x509_cert_dir, sc->x509_cert_file,
	     sc->x509_user_cert, sc->x509_user_key, x509_user_proxy, 
             tkt_ccache, NULL) < 0)
	perror("exec failed\n");
  } else { /* parent */
    fprintf(stderr, "%d: parent process waiting for child\n", getpid());
    waitpid(pid, &status, 0);
    fprintf(stderr, "%d: child is done, exit status= %d\n", pid, status);
  }
  if(!status && r->method_number == M_GET) return OK;
  else return DECLINED;	
  
}

static command_rec grid_config_cmds[] = {
  {"GridProxyInit", grid_cmd_GridProxyInit, NULL, RSRC_CONF, TAKE1,
   "specify the filename of the grid_proxy_init executable"},
  {"X509_CERT_DIR", grid_cmd_X509CertDir, NULL, RSRC_CONF, TAKE1,
   "specify the path for X509_CERT_DIR"},
  {"X509_CERT_FILE", grid_cmd_X509CertFile, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_CERT_FILE"},
  {"X509_USER_CERT", grid_cmd_X509UserCert, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_USER_CERT"},
  {"X509_USER_KEY", grid_cmd_X509UserKey, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_USER_KEY"},
  {"X509_USER_PROXY", grid_cmd_X509UserProxy, NULL, RSRC_CONF, TAKE1,
   "specify the prefix for X509_USER_PROXY"},
  {"TktCachePrefix", grid_cmd_GridTktCache, NULL, RSRC_CONF, TAKE1,
   "specify the prefix for Kerberos ticket cache"},
  {NULL},
}; 
 
static handler_rec grid_proxy_init_handlers[] = {
  {"*/*", grid_proxy_init_handler},
  {NULL}
};

module MODULE_VAR_EXPORT grid_proxy_init_module = {
  STANDARD_MODULE_STUFF,
  NULL, /* initializer */
  NULL, /* create per-dir config */
  NULL, /* merge per-dir config */
  grid_config_server_create, /* create per-server config */
  grid_config_server_merge, /* merge per-server config */
  grid_config_cmds, /* command table */
  grid_proxy_init_handlers, /* [#8] handlers */
  NULL, /* [#1] file name translation */
  NULL, /* [#4] check user id */
  NULL, /* [#5] check auth */
  NULL, /* [#3] check access */
  NULL, /* [#6] type checker */ 
  NULL, /* [#7] prerun fixups */ 
  NULL, /* [#9] logger */
  NULL, /* [#2] header parser */
  NULL, /* child init */
  NULL, /* child exit */
  NULL,  /* [#0] post read request */
};
